"""This is the API documentation for risus-py.

The library has several submodules:

The module `risus.engine` implements the computatations of probabilities in the Risus game system by S. John Ross.

The module `risus.table` produces tables of values based on computations done by `risus.engine`.

The module `risus.damage_policy` contains functions to distribute damage in team combat.

"""
